@@ -45,14 +45,6 @@ USE OR OTHER DEALINGS IN THE SOFTWARE.
 #include <stdio.h>
 #include <errno.h>
 
-#include <xf86cmap.h>
-#include <xf86drm.h>
-#include <xf86RandR12.h>
-#include <mi.h>
-#include <micmap.h>
-#include <mipict.h>
-
-#include "compiler.h"
 #include "sna.h"
 #include "sna_module.h"
 #include "sna_video.h"
@@ -60,6 +52,13 @@ USE OR OTHER DEALINGS IN THE SOFTWARE.
 #include "intel_driver.h"
 #include "intel_options.h"
 
+#include <xf86cmap.h>
+#include <xf86drm.h>
+#include <xf86RandR12.h>
+#include <mi.h>
+#include <micmap.h>
+#include <mipict.h>
+
 #include <sys/ioctl.h>
 #include <sys/fcntl.h>
 #include "i915_drm.h"
@@ -294,12 +293,20 @@ static void sna_selftest(void)
 	sna_damage_selftest();
 }
 
+static bool has_vsync(struct sna *sna)
+{
+	if (sna->flags & SNA_IS_HOSTED)
+		return false;
+
+	return true;
+}
+
 static bool has_pageflipping(struct sna *sna)
 {
 	drm_i915_getparam_t gp;
 	int v;
 
-	if (sna->flags & (SNA_IS_HOSTED | SNA_NO_WAIT))
+	if (sna->flags & SNA_IS_HOSTED)
 		return false;
 
 	v = 0;
@@ -330,9 +337,73 @@ static void sna_setup_capabilities(ScrnInfoPtr scrn, int fd)
 #endif
 }
 
+static int
+namecmp(const char *s1, const char *s2)
+{
+	char c1, c2;
+
+	if (!s1 || *s1 == 0) {
+		if (!s2 || *s2 == 0)
+			return 0;
+		else
+			return 1;
+	}
+
+	while (*s1 == '_' || *s1 == ' ' || *s1 == '\t')
+		s1++;
+
+	while (*s2 == '_' || *s2 == ' ' || *s2 == '\t')
+		s2++;
+
+	c1 = isupper(*s1) ? tolower(*s1) : *s1;
+	c2 = isupper(*s2) ? tolower(*s2) : *s2;
+	while (c1 == c2) {
+		if (c1 == '\0')
+			return 0;
+
+		s1++;
+		while (*s1 == '_' || *s1 == ' ' || *s1 == '\t')
+			s1++;
+
+		s2++;
+		while (*s2 == '_' || *s2 == ' ' || *s2 == '\t')
+			s2++;
+
+		c1 = isupper(*s1) ? tolower(*s1) : *s1;
+		c2 = isupper(*s2) ? tolower(*s2) : *s2;
+	}
+
+	return c1 - c2;
+}
+
 static Bool sna_option_cast_to_bool(struct sna *sna, int id, Bool val)
 {
-	xf86getBoolValue(&val, xf86GetOptValString(sna->Options, id));
+	const char *str = xf86GetOptValString(sna->Options, id);
+
+	if (str == NULL)
+		return val;
+
+	if (*str == '\0')
+		return TRUE;
+
+	if (namecmp(str, "1") == 0)
+		return TRUE;
+	if (namecmp(str, "on") == 0)
+		return TRUE;
+	if (namecmp(str, "true") == 0)
+		return TRUE;
+	if (namecmp(str, "yes") == 0)
+		return TRUE;
+
+	if (namecmp(str, "0") == 0)
+		return FALSE;
+	if (namecmp(str, "off") == 0)
+		return FALSE;
+	if (namecmp(str, "false") == 0)
+		return FALSE;
+	if (namecmp(str, "no") == 0)
+		return FALSE;
+
 	return val;
 }
 
@@ -430,8 +501,6 @@ static Bool sna_pre_init(ScrnInfoPtr scrn, int flags)
 
 	scrn->displayWidth = 640;	/* default it */
 
-	sna->PciInfo = xf86GetPciInfoForEntity(sna->pEnt->index);
-
 	scrn->monitor = scrn->confScreen->monitor;
 	scrn->progClock = TRUE;
 	scrn->rgbBits = 8;
@@ -487,9 +556,11 @@ static Bool sna_pre_init(ScrnInfoPtr scrn, int flags)
 
 	sna_setup_capabilities(scrn, fd);
 
-	intel_detect_chipset(scrn, sna->pEnt, sna->PciInfo);
+	intel_detect_chipset(scrn, pEnt);
 
-	kgem_init(&sna->kgem, fd, sna->PciInfo, sna->info->gen);
+	kgem_init(&sna->kgem, fd,
+		  xf86GetPciInfoForEntity(pEnt->index),
+		  sna->info->gen);
 	if (xf86ReturnOptValBool(sna->Options, OPTION_ACCEL_DISABLE, FALSE) ||
 	    !sna_option_cast_to_bool(sna, OPTION_ACCEL_METHOD, TRUE)) {
 		xf86DrvMsg(sna->scrn->scrnIndex, X_CONFIG,
@@ -508,13 +579,27 @@ static Bool sna_pre_init(ScrnInfoPtr scrn, int flags)
 
 	if (!xf86ReturnOptValBool(sna->Options, OPTION_SWAPBUFFERS_WAIT, TRUE))
 		sna->flags |= SNA_NO_WAIT;
-	if (xf86ReturnOptValBool(sna->Options, OPTION_TRIPLE_BUFFER, TRUE))
-		sna->flags |= SNA_TRIPLE_BUFFER;
-	if (has_pageflipping(sna)) {
-		if (xf86ReturnOptValBool(sna->Options, OPTION_TEAR_FREE, FALSE))
-			sna->flags |= SNA_TEAR_FREE;
-	} else
+	DBG(("%s: swapbuffer wait? %s\n", __FUNCTION__, sna->flags & SNA_NO_WAIT ? "disabled" : "enabled"));
+
+	if (!has_vsync(sna) ||
+	    !xf86ReturnOptValBool(sna->Options, OPTION_VSYNC, TRUE))
+		sna->flags |= SNA_NO_VSYNC;
+	DBG(("%s: vsync? %s\n", __FUNCTION__, sna->flags & SNA_NO_VSYNC ? "disabled" : "enabled"));
+
+	if (!has_pageflipping(sna) ||
+	    !xf86ReturnOptValBool(sna->Options, OPTION_PAGEFLIP, TRUE))
 		sna->flags |= SNA_NO_FLIP;
+	DBG(("%s: page flips? %s\n", __FUNCTION__, sna->flags & SNA_NO_FLIP ? "disabled" : "enabled"));
+
+	if ((sna->flags & (SNA_NO_VSYNC | SNA_NO_FLIP | SNA_NO_WAIT)) == 0 &&
+	    xf86ReturnOptValBool(sna->Options, OPTION_TRIPLE_BUFFER, TRUE))
+		sna->flags |= SNA_TRIPLE_BUFFER;
+	DBG(("%s: triple buffer? %s\n", __FUNCTION__, sna->flags & SNA_TRIPLE_BUFFER ? "enabled" : "disabled"));
+
+	if ((sna->flags & (SNA_NO_VSYNC | SNA_NO_FLIP)) == 0 &&
+	    xf86ReturnOptValBool(sna->Options, OPTION_TEAR_FREE, FALSE))
+		sna->flags |= SNA_TEAR_FREE;
+
 	if (xf86ReturnOptValBool(sna->Options, OPTION_CRTC_PIXMAPS, FALSE))
 		sna->flags |= SNA_FORCE_SHADOW;
 
@@ -739,6 +824,8 @@ static void sna_leave_vt(VT_FUNC_ARGS_DECL)
 
 	xf86_hide_cursors(scrn);
 
+	sna_mode_reset(to_sna(scrn));
+
 	if (intel_put_master(scrn))
 		xf86DrvMsg(scrn->scrnIndex, X_WARNING,
 			   "drmDropMaster failed: %s\n", strerror(errno));
@@ -852,6 +939,7 @@ sna_screen_init(SCREEN_INIT_ARGS_DECL)
 {
 	ScrnInfoPtr scrn = xf86ScreenToScrn(screen);
 	struct sna *sna = to_sna(scrn);
+	struct pci_device *pci;
 	VisualPtr visuals;
 	DepthPtr depths;
 	int nvisuals;
@@ -869,7 +957,11 @@ sna_screen_init(SCREEN_INIT_ARGS_DECL)
 	if (!sna_register_all_privates())
 		return FALSE;
 
-	scrn->videoRam = agp_aperture_size(sna->PciInfo, sna->kgem.gen) / 1024;
+	pci = xf86GetPciInfoForEntity(sna->pEnt->index);
+	if (pci != NULL)
+		scrn->videoRam = agp_aperture_size(pci, sna->kgem.gen) / 1024;
+	else
+		scrn->videoRam = 256;
 
 	miClearVisualTypes();
 	if (!miSetVisualTypes(scrn->depth,
