@@ -22,17 +22,27 @@
  *
  */
 
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
 #include <X11/Xlib.h>
 #include <X11/Xatom.h>
 
 #include <X11/Xlibint.h>
 #include <X11/extensions/record.h>
+#include <X11/extensions/XShm.h>
+#if HAVE_X11_EXTENSIONS_SHMPROTO_H
 #include <X11/extensions/shmproto.h>
+#elif HAVE_X11_EXTENSIONS_SHMSTR_H
+#include <X11/extensions/shmstr.h>
+#else
+#error Failed to find the right header for X11 MIT-SHM protocol definitions
+#endif
 #include <X11/extensions/Xdamage.h>
 #include <X11/extensions/Xinerama.h>
 #include <X11/extensions/Xrandr.h>
 #include <X11/extensions/Xrender.h>
-#include <X11/extensions/XShm.h>
 #include <X11/Xcursor/Xcursor.h>
 #include <pixman.h>
 
@@ -62,6 +72,8 @@
 #define DBG(x)
 #endif
 
+#define FORCE_FULL_REDRAW 0
+
 struct display {
 	Display *dpy;
 	struct clone *clone;
@@ -280,7 +292,10 @@ static int timerfd(int hz)
 	struct itimerspec it;
 	int fd;
 
+	fd = -1;
+#ifdef CLOCK_MONOTONIC_COARSE
 	fd = timerfd_create(CLOCK_MONOTONIC_COARSE, TFD_NONBLOCK);
+#endif
 	if (fd < 0)
 		fd = timerfd_create(CLOCK_MONOTONIC, TFD_NONBLOCK);
 	if (fd < 0)
@@ -397,8 +412,8 @@ static XRRModeInfo *lookup_mode(XRRScreenResources *res, int id)
 
 static int clone_update_modes__randr(struct clone *clone)
 {
-	XRRScreenResources *from_res, *to_res;
-	XRROutputInfo *from_info, *to_info;
+	XRRScreenResources *from_res = NULL, *to_res = NULL;
+	XRROutputInfo *from_info = NULL, *to_info = NULL;
 	int i, j, ret = ENOENT;
 
 	assert(clone->src.rr_output);
@@ -439,6 +454,7 @@ static int clone_update_modes__randr(struct clone *clone)
 	clone->src.rr_crtc = 0;
 
 	/* Create matching modes for the real output on the virtual */
+	XGrabServer(clone->src.dpy);
 	for (i = 0; i < from_info->nmode; i++) {
 		XRRModeInfo *mode, *old;
 		RRMode id;
@@ -479,6 +495,7 @@ static int clone_update_modes__randr(struct clone *clone)
 
 		XRRAddOutputMode(clone->src.dpy, clone->src.rr_output, id);
 	}
+	XUngrabServer(clone->src.dpy);
 	ret = 0;
 
 err:
@@ -497,8 +514,8 @@ err:
 static int clone_update_modes__fixed(struct clone *clone)
 {
 	char mode_name[80];
-	XRRScreenResources *res;
-	XRROutputInfo *info;
+	XRRScreenResources *res = NULL;
+	XRROutputInfo *info = NULL;
 	XRRModeInfo mode;
 	RRMode id;
 	int i, j, ret = ENOENT;
@@ -513,6 +530,8 @@ static int clone_update_modes__fixed(struct clone *clone)
 	if (info == NULL)
 		goto err;
 
+	XGrabServer(clone->src.dpy);
+
 	/* Clear all current UserModes on the output, including any active ones */
 	if (info->crtc) {
 		DBG(("%s(%s-%s): disabling active CRTC\n", __func__,
@@ -547,12 +566,14 @@ static int clone_update_modes__fixed(struct clone *clone)
 
 	XRRAddOutputMode(clone->src.dpy, clone->src.rr_output, id);
 
+	XUngrabServer(clone->src.dpy);
 	ret = 0;
 err:
 	if (info)
 		XRRFreeOutputInfo(info);
 	if (res)
 		XRRFreeScreenResources(res);
+
 	return ret;
 }
 
@@ -564,14 +585,15 @@ static RROutput claim_virtual(struct display *display, char *output_name, int nc
 	XRROutputInfo *output;
 	XRRModeInfo mode;
 	RRMode id;
-	RROutput rr_output;
+	RROutput rr_output = 0;
 	int i;
 
 	DBG(("%s(%d)\n", __func__, nclone));
+	XGrabServer(dpy);
 
 	res = _XRRGetScreenResourcesCurrent(dpy, display->root);
 	if (res == NULL)
-		return 0;
+		goto out;
 
 	sprintf(output_name, "VIRTUAL%d", nclone);
 
@@ -593,7 +615,7 @@ static RROutput claim_virtual(struct display *display, char *output_name, int nc
 
 	DBG(("%s(%s): rr_output=%ld\n", __func__, output_name, (long)rr_output));
 	if (rr_output == 0)
-		return 0;
+		goto out;
 
 	/* Set any mode on the VirtualHead to make the Xserver allocate another */
 	memset(&mode, 0, sizeof(mode));
@@ -609,7 +631,7 @@ static RROutput claim_virtual(struct display *display, char *output_name, int nc
 	/* Force a redetection for the ddx to spot the new outputs */
 	res = XRRGetScreenResources(dpy, display->root);
 	if (res == NULL)
-		return 0;
+		goto out;
 
 	/* Some else may have interrupted us and installed that new mode! */
 	output = XRRGetOutputInfo(dpy, res, rr_output);
@@ -624,6 +646,9 @@ static RROutput claim_virtual(struct display *display, char *output_name, int nc
 	XRRDeleteOutputMode(dpy, rr_output, id);
 	XRRDestroyMode(dpy, id);
 
+out:
+	XUngrabServer(dpy);
+
 	return rr_output;
 }
 
@@ -824,7 +849,7 @@ static void context_update(struct context *ctx)
 			c = XRRGetCrtcInfo(dpy, res, o->crtc);
 		if (c) {
 			DBG(("%s-%s: (x=%d, y=%d, rotation=%d, mode=%ld) -> (x=%d, y=%d, rotation=%d, mode=%ld)\n",
-			     DisplayString(ctx->display->dpy), output->name,
+			     DisplayString(dpy), output->name,
 			     output->x, output->y, output->rotation, output->mode.id,
 			     c->x, c->y, output->rotation, c->mode));
 
@@ -842,7 +867,7 @@ static void context_update(struct context *ctx)
 			XRRFreeCrtcInfo(c);
 		} else {
 			DBG(("%s-%s: (x=%d, y=%d, rotation=%d, mode=%ld) -> off\n",
-			     DisplayString(ctx->display->dpy), output->name,
+			     DisplayString(dpy), output->name,
 			     output->x, output->y, output->rotation, output->mode.id));
 		}
 		output->rr_crtc = o->crtc;
@@ -871,7 +896,7 @@ static void context_update(struct context *ctx)
 	}
 	XRRFreeScreenResources(res);
 
-	DBG(("%s changed? %d\n", DisplayString(ctx->display->dpy), context_changed));
+	DBG(("%s changed? %d\n", DisplayString(dpy), context_changed));
 	if (!context_changed)
 		return;
 
@@ -920,6 +945,7 @@ static void context_update(struct context *ctx)
 		if (res == NULL)
 			continue;
 
+		XGrabServer(display->dpy);
 		for (clone = display->clone; clone; clone = clone->next) {
 			struct output *src = &clone->src;
 			struct output *dst = &clone->dst;
@@ -1005,6 +1031,7 @@ err:
 					 &dst->rr_output, 1);
 			dst->rr_crtc = rr_crtc;
 		}
+		XUngrabServer(display->dpy);
 
 		XRRFreeScreenResources(res);
 	}
@@ -1329,6 +1356,13 @@ static int clone_paint(struct clone *c)
 	if (c->dst.serial > LastKnownRequestProcessed(c->dst.dpy))
 		return EAGAIN;
 
+	if (FORCE_FULL_REDRAW) {
+		c->damaged.x1 = c->src.x;
+		c->damaged.y1 = c->src.y;
+		c->damaged.x2 = c->src.x + c->width;
+		c->damaged.y2 = c->src.y + c->height;
+	}
+
 	clip.x = c->damaged.x1;
 	clip.y = c->damaged.y1;
 	clip.width  = c->damaged.x2 - c->damaged.x1;
@@ -1595,6 +1629,7 @@ static int clone_init_depth(struct clone *clone)
 	DBG(("%s-%s wants depth %d\n",
 	     DisplayString(clone->dst.dpy), clone->dst.name, clone->depth));
 
+	ret = -1;
 	for (depth = clone->depth; depth <= 24; depth += 8) {
 		ret = display_init_render(clone->src.display, depth, &clone->src.use_render);
 		if (ret)
@@ -1663,6 +1698,8 @@ static int add_display(struct context *ctx, Display *dpy)
 	display->invisible_cursor = display_load_invisible_cursor(display);
 	display->cursor = None;
 
+	display_cursor_move(display, 0, 0, 0);
+
 	return ConnectionNumber(dpy);
 }
 
@@ -1977,7 +2014,7 @@ static int last_display_add_clones__display(struct context *ctx)
 	Display *dpy = display->dpy;
 	Screen *scr;
 	char buf[80];
-	int n, ret;
+	int ret;
 	RROutput id;
 
 	DBG(("%s(%s)\n", __func__, DisplayString(display->dpy)));
@@ -1998,7 +2035,7 @@ static int last_display_add_clones__display(struct context *ctx)
 		return ret;
 	}
 
-	sprintf(buf, "WHOLE%d", n);
+	sprintf(buf, "WHOLE");
 	ret = clone_output_init(clone, &clone->dst, display, buf, 0);
 	if (ret) {
 		fprintf(stderr, "Failed to add display \"%s\"\n",
@@ -2352,6 +2389,8 @@ static void context_cleanup(struct context *ctx)
 	if (res == NULL)
 		return;
 
+	XGrabServer(dpy);
+
 	for (i = 0; i < ctx->nclone; i++) {
 		struct clone *clone = &ctx->clones[i];
 		XRROutputInfo *output;
@@ -2384,6 +2423,7 @@ static void context_cleanup(struct context *ctx)
 		}
 	}
 
+	XUngrabServer(dpy);
 	XRRFreeScreenResources(res);
 
 	if (ctx->singleton)
